{-# LANGUAGE DeriveGeneric #-}

import           Control.Arrow ((&&&), second)
import           Control.Monad (liftM, when)
import           Data.Foldable (toList)
import qualified Data.ByteString.Char8 as BC
import qualified Data.IntMap as IM
import qualified Data.IntSet as IS
import           Data.List (findIndices, inits, intercalate, sortBy, transpose)
import           Data.Maybe (mapMaybe)
import           Data.Ord (comparing)
import qualified Data.Sequence as Seq
import           GHC.Generics (Generic)
import           System.Console.GetOpt
import           System.Environment (getArgs, getProgName)
import           System.Exit (exitFailure)
import           System.IO (hPutStrLn, stderr)
import           Text.Printf (printf)

import qualified Interpreter as I
import           Table
import           Types
import           Utils


data SampleStatistics = SampleStatistics {
  auc       :: Double
, cover     :: Double
, rank      :: Double
, precision :: Int
, recall    :: Int
, coverI    :: [Double]
, recallI   :: [Int]
, lblRanks  :: [(LabelID, Maybe Int)]
, sampleID  :: SampleID
} deriving Generic

data FileStatistics = FileStatistics {
  aucF       :: Double
, coverF     :: Double
, rankF      :: Double
, precisionF :: Double
, recallF    :: Double
, coverIF    :: [Double]
, recallIF   :: [Int]
, lblRanksF  :: [(SampleID, [(LabelID, Maybe Int)])]
, nSamples   :: Int
}


-- -----------------------------------------------------------------------------
-- Sample statistics

calculateAUC :: Int -> Int -> [Int] -> Double
calculateAUC nLabels nPreds hitIndices =
  if nLabels == 0 || neg == 0 then 1
  else fromIntegral (sum areas) / fromIntegral (nLabels * neg)
  where neg = nPreds - length hitIndices
        areas = zipWith3 (\ l h n -> (h - l) * n) (init hIs) (tail hIs) [1 ..]
        hIs = hitIndices ++ [nPreds]

calculateRank :: Int -> Int -> [Int] -> Double
calculateRank nDeps nPreds hitIndices =
  fromIntegral (sum hitIndices + missed * nPreds) / fromIntegral nDeps
  where missed = nDeps - length hitIndices

calculateRecall :: Int -> Int -> [Int] -> Int
calculateRecall nDeps nPreds hitIndices
  | length hitIndices == nDeps = 1 + last hitIndices
  | otherwise                  = 1 + nPreds

calculateCoverI :: Int -> [Int] -> [Int] -> [Double]
calculateCoverI nLabels hitIndices =
  map (\ i -> missed i / fromIntegral (min nLabels i))
  where missed i = fromIntegral $ length (takeWhile (< i) hitIndices)


-- | Compare actual labels with predicted labels.
evaluateSample :: SampleID -> Labels -> Labels -> Maybe SampleStatistics
evaluateSample _ []   _     = Nothing
evaluateSample _ _    []    = Nothing
evaluateSample s lbls preds = Just SampleStatistics {
    auc       = calculateAUC nLabels nPreds hitIndices
  , cover     = fromIntegral (length lblsInPreds) / nLabelsN
  , rank      = calculateRank nLabels nPreds hitIndices
  , precision = length predsInLbls
  , recall    = recall'
  , coverI    = calculateCoverI nLabels hitIndices indices
  , recallI   = map (\ i -> if i >= recall' then 1 else 0) indices
  , lblRanks  = map (\ l -> (l, IM.lookup l preds100M)) lbls
  , sampleID  = s
  }
  where
    nLabels   = length lbls
    nPreds    = length preds
    nLabelsN  = fromIntegral nLabels
    preds100  = take 100 preds
    preds100M = IM.fromList $ zip preds100 [1 ..]
    lblsS     = IS.fromList lbls

    lblsInPreds = filter (`IM.member` preds100M) lbls
    predsInLbls = filter (`IS.member` lblsS) preds100
    hitIndices  = findIndices (`IS.member` lblsS) preds

    indices = [1 .. 200]
    --indices = [2^x | x <- [0 .. 11]]

    recall'  = calculateRecall nLabels nPreds hitIndices


-- -----------------------------------------------------------------------------
-- File statistics

evaluateSamples :: [(SampleID, (Labels, Labels))] -> [SampleStatistics]
evaluateSamples = mapMaybe (\ (s, (l, p)) -> evaluateSample s l p)

-- | Match labels with predictions.
match :: LabelTable -> LabelTable -> [(SampleID, (Labels, Labels))]
match lbls = toList . Seq.mapWithIndex (\ s lp -> (SampleID s, lp)) . Seq.zip lbls

combineStats :: [SampleStatistics] -> FileStatistics
combineStats stats = FileStatistics {
    aucF       = sum (map auc       stats) / n
  , coverF     = sum (map cover     stats) / n
  , rankF      = sum (map rank      stats) / n
  , precisionF = fromIntegral (sum (map precision stats)) / n
  , recallF    = fromIntegral (sum (map recall    stats)) / n
  , coverIF    = map (100 / n *) $ sumLists $ map coverI stats
  , recallIF   = sumLists $ map recallI stats
  , lblRanksF  = map (sampleID &&& lblRanks) stats
  , nSamples   = nI
  }
  where nI    = length stats
        n     = fromIntegral nI

-- | Calculate statistics for increasing number of samples.
iterativeStats :: [SampleStatistics] -> [FileStatistics]
iterativeStats = map combineStats . tail . inits


-- -----------------------------------------------------------------------------
-- Printing

statisticsList :: FilePath -> FileStatistics -> [String]
statisticsList fp stats =
  fp : show (nSamples stats) : zipWith (\ f d -> printf ("%." ++ show (d :: Int) ++ "f") $ f stats)
  [coverF, precisionF, recallF, aucF, rankF]
  [     3,          3,       2,    4,     2]

showLblRanks :: LabelMap -> FileStatistics -> [String]
showLblRanks lblMap =
  map (\ (s, r) -> unlines $ (showSampleS s ++ ":") : showRanks r) . lblRanksF
  where
    showSampleS = BC.unpack . showSample lblMap
    showLabelS = BC.unpack . showLabel lblMap
    showRanks = map (uncurry showRank) . sortBy (comparing snd)
    showRank lbl rnk = "  " ++
      maybe "N/A" (printf "%3d") rnk ++ " -> " ++ showLabelS lbl


-- -----------------------------------------------------------------------------
-- Command-line options

data Options = Options {
  optCSV     :: Bool
, optCoverI  :: Bool
, optRecallI :: Bool
, optRanks   :: Bool
, optIterate :: Bool
}

defaultOptions :: Options
defaultOptions = Options {
  optCSV     = False
, optCoverI  = False
, optRecallI = False
, optRanks   = False
, optIterate = False
}

options :: [OptDescr (Options -> Options)]
options =
  [ Option "t" []
      (NoArg (\ opts -> opts {optCSV = True})) "output tables in CSV format"
  , Option "c" []
      (NoArg (\ opts -> opts {optCoverI  = True})) "print Cover-i"
  , Option "r" []
      (NoArg (\ opts -> opts {optRecallI = True})) "print Recall-i"
  , Option "p" []
      (NoArg (\ opts -> opts {optRanks = True})) "print prediction ranks"
  , Option "i" []
      (NoArg (\ opts -> opts {optIterate = True})) "iterative statistics"
  ]


-- -----------------------------------------------------------------------------
-- Help

header :: String -> String
header progName = intercalate "\n"
  [ "Usage: " ++ progName ++ " <seq> <deps> [OPTS] [<preds1> [... [<preds_n>]]]"
  , ""
  , "Mandatory arguments:"
  , "  <seq>    file containing the order of theorems"
  , "  <deps>   file containing the dependencies of theorems"
  , ""
  , "Optional arguments:"
  ]

showUsage :: IO ()
showUsage = do
  progName <- getProgName
  hPutStrLn stderr $ usageInfo (header progName) options


-- -----------------------------------------------------------------------------
-- Main

-- | Load predictions either from files or from stdin if file list is empty.
loadPredictions :: LabelMap -> [FilePath] -> IO [(FilePath, LabelTable)]
loadPredictions lblMap predFPs =
  if null predFPs
  then liftM (\ c -> [("<stdin>", I.loadLabels lblMap $ BC.lines c)]) BC.getContents
  else mapM (\ fp -> liftPair (return fp, I.loadLabelsFile lblMap fp)) predFPs


main :: IO ()
main = do
  args <- getArgs
  (seqFP, labelsFP, predFPs, opts) <- case getOpt Permute options args of
    (o, sF : lF : pFs, []) ->
      return (sF, lF, pFs, foldl (flip id) defaultOptions o)
    _ -> showUsage >> exitFailure

  lblMap <- I.loadSequenceFile seqFP
  lbls   <- I.loadLabelsFile lblMap labelsFP
  preds  <- loadPredictions lblMap predFPs

  let sampleStats = map (second $ evaluateSamples . match lbls) preds
  let fileStats = (if optIterate opts then iterStats else allStats) sampleStats

  let descriptors = ["", "Samples", "Cov", "Prec", "Rec", "AUC", "Rank"]
      statisticsOpts = TableOptions {
        separator = if optCSV opts then ";" else " "
      , cellWidths = maximum (map (length . fst) preds) : repeat 7
      , alignments = AlignLeft : repeat AlignRight
      }
      graphOpts = statisticsOpts {
        cellWidths = map (length . fst) preds
      , alignments = repeat AlignRight
      }

      stats = descriptors : map (uncurry statisticsList) fileStats
      covIs = map (\ (fp, s) -> fp : show2fs  (coverIF  s)) fileStats
      recIs = map (\ (fp, s) -> fp : map show (recallIF s)) fileStats
      lRnks = map (\ (fp, s) -> fp : showLblRanks lblMap s) fileStats

  let showStats = not $ optCoverI opts || optRecallI opts || optRanks opts

  when showStats $ putStr $ renderTable statisticsOpts stats
  when (optCoverI  opts) $ putStr $ renderTable graphOpts $ transpose covIs
  when (optRecallI opts) $ putStr $ renderTable graphOpts $ transpose recIs
  when (optRanks   opts) $ mapM_ (putStrLn . unlines) lRnks

  where iterStats = concatMap (\ (fp, p) -> zip (repeat fp) $ iterativeStats p)
        allStats = map (second combineStats)
        show2fs = map (printf "%.2f")
